﻿using System.Collections.Generic;

namespace NeuroExplorer.NexFile
{
    /// <summary>
    /// Continuous variable.
    /// </summary>
    /// <remarks>
    /// In general, a continuous variable may contain several fragments of data.
    /// Each fragment may be of different length. We don't store the timestamps
    /// for all the a/d values since they would use too much space. Instead,
    /// for each fragment we store the timestamp of the first a/d value in
    /// the fragment and the index of the first data point in the fragment.
    /// Therefore, a continuous variable contains the following 3 arrays:
    /// 
    /// - array of all a/d values in milliVolts
    /// NexVarHeader.NPointsWave field stores the number of a/d values
    /// 
    /// - array of timestamps
    /// each timestamp is for the beginning of the fragment;
    /// NexVarHeader.Count field stores the number of fragments
    /// 
    /// - array of indexes 
    /// each index is the position of the first data point
    /// of the fragment in the a/d array; index[0] is always 0, 
    /// if index[1] = 200, it means that the second fragment is advalues[200], advalues[201], etc.)
    ///</remarks>

    public class Continuous : Variable
    {
        /// <summary>
        /// Array of timestamps in seconds (each timestamp is for the beginning of the fragment).
        /// </summary>
        public List<double> FragmentStartsTimestamps { get; set; }

        /// <summary>
        /// Array of indexes (each index is the position of the first data point.
        /// of the fragment in the a/d array).
        /// </summary>
        public List<uint> FragmentStartsIndexes { get; set; }

        /// <summary>
        /// Array of all a/d values in milliVolts.
        /// </summary>
        public List<float> Values { get; set; }

        /// <summary>
        /// Sampling rate in Hz.
        /// </summary>
        public double SamplingRate { get; set; }

        public double CalulatedCoeffToScaleFloatsToShorts { get; set; }

        public Continuous() : base()
        {
            FragmentStartsTimestamps = new List<double>();
            FragmentStartsIndexes = new List<uint>();
            Values = new List<float>();
            SamplingRate = 0;
            CalulatedCoeffToScaleFloatsToShorts = 1;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="name"></param>
        /// <param name="samplingRate"></param>
        public Continuous( string name, double samplingRate )
        {
            Name = name;
            FragmentStartsTimestamps = new List<double>();
            FragmentStartsIndexes = new List<uint>();
            Values = new List<float>();
            SamplingRate = samplingRate;
            CalulatedCoeffToScaleFloatsToShorts = 1;
        }

        /// <summary>
        /// Constructor for a simple cont. var with a single fragment
        /// </summary>
        /// <param name="name"></param>
        /// <param name="samplingRate"></param>
        public Continuous( string name, double samplingRate, double startTime, float[] values )
        {
            Name = name;
            FragmentStartsTimestamps = new List<double>();
            FragmentStartsTimestamps.Add( startTime );
            FragmentStartsIndexes = new List<uint>();
            FragmentStartsIndexes.Add( 0 );
            Values = new List<float>(values);
            SamplingRate = samplingRate;
            CalulatedCoeffToScaleFloatsToShorts = 1;
        }


        public void AddFragment( double startTimestamp, float[] values )
        {
            FragmentStartsTimestamps.Add( startTimestamp );
            FragmentStartsIndexes.Add( ( uint )Values.Count );
            Values.AddRange( values );
        }

        public override double MaxTimestamp()
        {
            int numFrags = FragmentStartsTimestamps.Count;
            if ( numFrags > 0 ) {
                double fragStart = FragmentStartsTimestamps[ numFrags - 1 ];
                double step = 1.0 / SamplingRate;
                return fragStart + step * ( Values.Count - FragmentStartsIndexes[ numFrags - 1 ] - 1 );
            } else {
                return 0;
            }
        }
    }
}
