﻿using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;

namespace NeuroExplorer.NexFile
{
    public class FileData
    {
        public double TimestampFrequency { get; set; }  // timestamps frequency, in Hertz
        public string Comment { get; set; } // file comment
        public double BegSeconds { get; set; } // minimum timestamp in seconds
        public double EndSeconds { get; set; } // maximum timestamp in seconds

        public List<Event> EventVariables { get; set; }
        public List<Neuron> NeuronVariables { get; set; }
        public List<Interval> IntervalVariables { get; set; }
        public List<Marker> MarkerVariables { get; set; }
        public List<Continuous> ContinuousVariables { get; set; }
        public List<Waveform> WaveformVariables { get; set; }

        public Dictionary<string, JToken> ExtraFileMetaData { get; set; }


        public FileData()
        {
            TimestampFrequency = 0;
            Comment = "";
            BegSeconds = 0;
            EndSeconds = 0;
            EventVariables = new List<Event>();
            NeuronVariables = new List<Neuron>();
            IntervalVariables = new List<Interval>();
            MarkerVariables = new List<Marker>();
            ContinuousVariables = new List<Continuous>();
            WaveformVariables = new List<Waveform>();
            ExtraFileMetaData = new Dictionary<string, JToken>();
        }

        public FileData( double tsFrequency, string comment )
        {
            TimestampFrequency = tsFrequency;
            Comment = comment;
            BegSeconds = 0;
            EndSeconds = 0;
            EventVariables = new List<Event>();
            NeuronVariables = new List<Neuron>();
            IntervalVariables = new List<Interval>();
            MarkerVariables = new List<Marker>();
            ContinuousVariables = new List<Continuous>();
            WaveformVariables = new List<Waveform>();
            ExtraFileMetaData = new Dictionary<string, JToken>();
        }

        public int NumberOfVariables()
        {
            return EventVariables.Count + NeuronVariables.Count + IntervalVariables.Count
                + MarkerVariables.Count + ContinuousVariables.Count + WaveformVariables.Count;
        }

        public long SecondsToLongTicks( double seconds )
        {
            return ( long )Math.Round( seconds * TimestampFrequency );
        }

        public int SecondsToIntTicks( double seconds )
        {
            return ( int )Math.Round( seconds * TimestampFrequency );
        }

        public List<long> SecondsToLongTicks( List<double> seconds )
        {
            List<long> ts = new List<long>();
            foreach ( var t in seconds ) {
                ts.Add( SecondsToLongTicks( t ) );
            }
            return ts;
        }

        public List<int> SecondsToIntTicks( List<double> seconds )
        {
            List<int> ts = new List<int>();
            foreach ( var t in seconds ) {
                ts.Add( SecondsToIntTicks( t ) );
            }
            return ts;
        }

        public double MaxTimestamp()
        {
            double maxTs = 0;
            foreach ( var v in NeuronVariables ) {
                maxTs = Math.Max( maxTs, v.MaxTimestamp() );
            }
            foreach ( var v in EventVariables ) {
                maxTs = Math.Max( maxTs, v.MaxTimestamp() );
            }
            foreach ( var v in IntervalVariables ) {
                maxTs = Math.Max( maxTs, v.MaxTimestamp() );
            }
            foreach ( var v in MarkerVariables ) {
                maxTs = Math.Max( maxTs, v.MaxTimestamp() );
            }
            foreach ( var v in ContinuousVariables ) {
                maxTs = Math.Max( maxTs, v.MaxTimestamp() );
            }
            foreach ( var v in WaveformVariables ) {
                maxTs = Math.Max( maxTs, v.MaxTimestamp() );
            }

            return maxTs;
        }
    }
}
