﻿using System;
using System.IO;
using System.Linq;
using System.Text;

// .nex5 file header structure
//struct Nex5FileHeader
//{
//    int Nex5MagicNumber; // string NEX5; numeric value: 894977358 decimal or 0x3558454e hex

//    int Nex5FileVersion;  // valid values are 500, 501 or 502 (April, 2017)
//                          // if 500, RecordingEndTimeInTicks is not specified
//                          // if 501 or greater, RecordingEndTimeInTicks is specified
//                          // if 502 or greater, timestamps are saved as 64-bit integers

//    char Comment[ 256 ]; // file comment; UTF-8 encoding is assumed.

//    double TimestampFrequency;  // timestamps frequency, in Hertz
//                                // timestamp values are stored in ticks, where tick = 1/Frequency.

//    long long RecordingStartTimeInTicks;    // when data is imported from a file with non-zero start time 

//    int NumberOfVariables;    // number of variables in the file, number of variables headers that follow this file header

//    unsigned long long MetadataOffset; // where the optional metadata for the file and its variables is located in the file
//                                       // metadata is a string in JSON format. see metadata description below. 

//    long long RecordingEndTimeInTicks;    // the maximum data timestamp across all file variables (file version 501 or greater)
//    // for neuron, event or marker variable, it is the last timestamp
//    // for interval variable, it is the timestamp of the end of the last interval
//    // for continuous variable, it is the timestamp of the last data point
//    // for waveform variable, it is the timestamp of the last data point of the last waveform

//    char Padding[ 56 ]; // padding for future expansion 
//};

namespace NeuroExplorer.NexFile
{
    public class Nex5FileHeader
    {
        public int Nex5MagicNumber { get; set; } // string NEX5; numeric value: 894977358 decimal or 0x3558454e hex
        public int Nex5FileVersion { get; set; }  // valid values are 500, 501 or 502 (April, 2017)
                                                  // if 500, RecordingEndTimeInTicks is not specified
                                                  // if 501 or greater, RecordingEndTimeInTicks is specified
                                                  // if 502 or greater, timestamps are saved as 64-bit integers
        public string Comment { get; set; } // file comment; UTF-8 encoding is assumed.
        public double TimestampFrequency { get; set; }  // timestamps frequency, in Hertz
                                                        // timestamp values are stored in ticks, where tick = 1/Frequency.
        public long RecordingStartTimeInTicks { get; set; }   // when data is imported from a file with non-zero start time 
        public int NumberOfVariables { get; set; }   // number of variables in the file, number of variables headers that follow this file header
        public ulong MetadataOffset { get; set; }// where the optional metadata for the file and its variables is located in the file
                                                 // metadata is a string in JSON format. see metadata description below. 
        public long RecordingEndTimeInTicks { get; set; }    // the maximum data timestamp across all file variables (file version 501 or greater)
                                                      // for neuron, event or marker variable, it is the last timestamp
                                                      // for interval variable, it is the timestamp of the end of the last interval
                                                      // for continuous variable, it is the timestamp of the last data point
                                                      // for waveform variable, it is the timestamp of the last data point of the last waveform
        public Nex5FileHeader()
        {
            Nex5MagicNumber = 0;
            Nex5FileVersion = 500;
            Comment = "";
            TimestampFrequency = 0.0;
            RecordingStartTimeInTicks = 0;
            NumberOfVariables = 0;
            MetadataOffset = 0;
            RecordingEndTimeInTicks = 0;
        }

        public void ReadFromFile( BinaryReader reader )
        {
            Nex5MagicNumber = reader.ReadInt32();
            if ( Nex5MagicNumber != 0x3558454e ) {
                throw new ApplicationException( "Invalid .nex5 file: magic number in the file header is incorrect" );
            }
            Nex5FileVersion = reader.ReadInt32();
            Comment = Encoding.UTF8.GetString( reader.ReadBytes( 256 ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() );
            TimestampFrequency = reader.ReadDouble();
            if ( TimestampFrequency <= 0 ) {
                throw new ApplicationException( "Invalid .nex5 file: timestamp frequency is not positive" );
            }
            RecordingStartTimeInTicks = reader.ReadInt64();
            NumberOfVariables = reader.ReadInt32();
            MetadataOffset = reader.ReadUInt64();
            RecordingEndTimeInTicks = reader.ReadInt64();
            var padding = reader.ReadBytes( 56 );
        }

        public void WriteToFile( BinaryWriter writer )
        {
            writer.Write( Nex5MagicNumber );
            writer.Write( Nex5FileVersion );
            writer.Write( Utilities.GetBytesFromStringUTF8( Comment, 256 ) );
            writer.Write( TimestampFrequency );
            writer.Write( RecordingStartTimeInTicks );
            writer.Write( NumberOfVariables );
            writer.Write( MetadataOffset );
            writer.Write( RecordingEndTimeInTicks );
            writer.Write( new byte[ 56 ] );
        }
    }
}
