﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Newtonsoft.Json;

namespace NeuroExplorer.NexFile
{
    public class Nex5FileReader
    {
        public FileData ReadNex5File( string filePath )
        {
            FileData fd = new FileData();
            using ( FileStream stream = new FileStream( filePath, FileMode.Open, FileAccess.Read, FileShare.Read ) ) {
                using ( BinaryReader reader = new BinaryReader( stream ) ) {
                    // 1. read file header
                    Nex5FileHeader fileHeader = new Nex5FileHeader();
                    fileHeader.ReadFromFile( reader );
                    fd.TimestampFrequency = fileHeader.TimestampFrequency;
                    fd.Comment = fileHeader.Comment;
                    fd.BegSeconds = fileHeader.RecordingStartTimeInTicks / fileHeader.TimestampFrequency;
                    fd.EndSeconds = fileHeader.RecordingEndTimeInTicks / fileHeader.TimestampFrequency;

                    // 2. read variable headers
                    List<Nex5FileVarHeader> nexVarHeaders = new List<Nex5FileVarHeader>();
                    for ( int i = 0; i < fileHeader.NumberOfVariables; i++ ) {
                        Nex5FileVarHeader varHeader = new Nex5FileVarHeader();
                        varHeader.ReadFromFile( reader );
                        nexVarHeaders.Add( varHeader );
                    }

                    // 3. read data
                    foreach ( var vh in nexVarHeaders ) {
                        switch ( vh.Type ) {
                            case VariableType.Event:
                                Event nexEvent = new Event();
                                nexEvent.Name = vh.Name;
                                reader.BaseStream.Seek( ( long )vh.DataOffset, SeekOrigin.Begin );
                                if ( vh.TimestampDataType == 0 ) {
                                    nexEvent.Timestamps = Utilities.ReadIntTimestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                } else {
                                    nexEvent.Timestamps = Utilities.ReadInt64Timestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                }
                                fd.EventVariables.Add( nexEvent );
                                break;

                            case VariableType.Neuron:
                                Neuron neuron = new Neuron();
                                neuron.Name = vh.Name;
                                reader.BaseStream.Seek( ( long )vh.DataOffset, SeekOrigin.Begin );
                                if ( vh.TimestampDataType == 0 ) {
                                    neuron.Timestamps = Utilities.ReadIntTimestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                } else {
                                    neuron.Timestamps = Utilities.ReadInt64Timestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                }
                                fd.NeuronVariables.Add( neuron );
                                break;

                            case VariableType.Interval:
                                Interval interval = new Interval();
                                interval.Name = vh.Name;
                                reader.BaseStream.Seek( ( long )vh.DataOffset, SeekOrigin.Begin );
                                if ( vh.TimestampDataType == 0 ) {
                                    interval.IntervalStarts = Utilities.ReadIntTimestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                    interval.IntervalEnds = Utilities.ReadIntTimestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                } else {
                                    interval.IntervalStarts = Utilities.ReadInt64Timestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                    interval.IntervalEnds = Utilities.ReadInt64Timestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                }
                                fd.IntervalVariables.Add( interval );
                                break;

                            case VariableType.Marker:
                                Marker marker = new Marker();
                                marker.Name = vh.Name;
                                reader.BaseStream.Seek( ( long )vh.DataOffset, SeekOrigin.Begin );
                                if ( vh.TimestampDataType == 0 ) {
                                    marker.Timestamps = Utilities.ReadIntTimestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                } else {
                                    marker.Timestamps = Utilities.ReadInt64Timestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                }
                                for ( int i = 0; i < vh.NumberOfMarkerFields; i++ ) {
                                    marker.MarkerNames.Add( Encoding.ASCII.GetString( reader.ReadBytes( 64 ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() ) );
                                    if ( vh.MarkerDataType == 0 ) {
                                        var s = new List<string>();
                                        for ( int j = 0; j < ( int )vh.Count; j++ ) {
                                            s.Add( Encoding.ASCII.GetString( reader.ReadBytes( vh.MarkerLength ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() ) );
                                        }
                                        marker.MarkerStrings.Add( s );
                                    } else {
                                        marker.MarkerValues.Add( Utilities.ReadUIntsArrayAsList( reader, ( int )vh.Count ) );
                                    }
                                }
                                fd.MarkerVariables.Add( marker );
                                break;

                            case VariableType.Continuous:
                                Continuous cont = new Continuous();
                                cont.Name = vh.Name;
                                cont.SamplingRate = vh.SamplingFrequency;
                                reader.BaseStream.Seek( ( long )vh.DataOffset, SeekOrigin.Begin );

                                if ( vh.TimestampDataType == 0 ) {
                                    cont.FragmentStartsTimestamps = Utilities.ReadIntTimestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                } else {
                                    cont.FragmentStartsTimestamps = Utilities.ReadInt64Timestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                }

                                if ( vh.ContinuousIndexOfFirstPointInFragmentDataType == 0 ) {
                                    cont.FragmentStartsIndexes = Utilities.ReadUIntsArrayAsList( reader, ( int )vh.Count );
                                } else {
                                    cont.FragmentStartsIndexes = Utilities.ReadUInts64ArrayAsListReturnUIntList( reader, ( int )vh.Count );
                                }

                                if ( vh.ContinuousDataType == 0 ) {
                                    cont.Values = Utilities.ReadAndScaleShortsArray( reader, ( int )vh.NumberOfDataPoints, vh.ADtoUnitsCoefficient, vh.UnitsOffset );
                                } else {
                                    cont.Values = Utilities.ReadFloatsArray( reader, ( int )vh.NumberOfDataPoints );
                                }

                                fd.ContinuousVariables.Add( cont );
                                break;

                            case VariableType.Waveform:
                                Waveform wave = new Waveform();
                                wave.Name = vh.Name;
                                wave.SamplingRate = vh.SamplingFrequency;
                                wave.NPointsWave = ( int )vh.NumberOfDataPoints;
                                wave.PrethresholdTimeInSeconds = vh.PrethresholdTimeInSeconds;
                                reader.BaseStream.Seek( ( long )vh.DataOffset, SeekOrigin.Begin );
                                if ( vh.TimestampDataType == 0 ) {
                                    wave.Timestamps = Utilities.ReadIntTimestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                } else {
                                    wave.Timestamps = Utilities.ReadInt64Timestamps( reader, ( int )vh.Count, fileHeader.TimestampFrequency );
                                }
                                if ( vh.ContinuousDataType == 0 ) {
                                    wave.Values = Utilities.ReadAndScaleShortsArrayWaves( reader, ( int )vh.Count, vh.ADtoUnitsCoefficient, vh.UnitsOffset, ( int )vh.NumberOfDataPoints );
                                } else {
                                    wave.Values = Utilities.ReadFloatArrayWaves( reader, ( int )vh.Count, ( int )vh.NumberOfDataPoints );
                                }
                                fd.WaveformVariables.Add( wave );
                                break;
                        }
                    }

                    // 4. read json 
                    var fileLength = reader.BaseStream.Length;
                    if ( fileHeader.MetadataOffset > 0 && ( long )fileHeader.MetadataOffset < fileLength ) {
                        reader.BaseStream.Seek( ( long )fileHeader.MetadataOffset, SeekOrigin.Begin );
                        long metaLength = fileLength - ( long )fileHeader.MetadataOffset;
                        var metaString = Encoding.UTF8.GetString( reader.ReadBytes( ( int )metaLength ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() );
                        try {
                            var meta = JsonConvert.DeserializeObject<Nex5Meta>( metaString );
                            if ( meta.variables != null ) {
                                foreach ( var varMeta in meta.variables ) {
                                    foreach ( var neuron in fd.NeuronVariables ) {
                                        if ( neuron.Name == varMeta.name ) {
                                            neuron.UnitNumber = varMeta.unitNumber;
                                            if ( varMeta.probe != null ) {
                                                neuron.WireNumber = varMeta.probe.wireNumber;
                                                if ( varMeta.probe.position != null ) {
                                                    neuron.XPos = varMeta.probe.position.x;
                                                    neuron.YPos = varMeta.probe.position.y;
                                                }
                                            }
                                            break;
                                        }
                                    }
                                    foreach ( var wave in fd.WaveformVariables ) {
                                        if ( wave.Name == varMeta.name ) {
                                            wave.UnitNumber = varMeta.unitNumber;
                                            if ( varMeta.probe != null ) {
                                                wave.WireNumber = varMeta.probe.wireNumber;
                                            }
                                            break;
                                        }
                                    }
                                }
                            }
                        } catch {
                        }
                    }
                } // reader
            }
            return fd;
        }
    }
}
