﻿using System.IO;
using System.Linq;
using System.Text;

namespace NeuroExplorer.NexFile
{
    /// <summary>
    /// VariableHeader is the header for each variable in the .nex5 data file.
    /// </summary>
    public class Nex5FileVarHeader
    {
        public VariableType Type = VariableType.Neuron; // 0 - neuron, 1 - event, 2- interval, 3 - waveform, 4 - pop. vector, 5 - continuously recorded, 6 - marker
                                                        // see #define statements above
        public int Version { get; set; } // Version = 500 as of July 2015 (version 5.016).
        public string Name { get; set; }// variable name. it is assumed that variable names are unique within the file. 
                                        // that is, there are no two variables with the same name.
                                        // there is one reserved name: AllFile 
                                        // NeuroExplorer creates AllFile variable from file data. 
                                        // It is an interval variable with a single interval [RecordingStartTimeInTicks, RecordingEndTimeInTicks]
        public ulong DataOffset { get; set; } // where the data array for this variable is located in the file
        public ulong Count { get; set; } // neuron variable: number of timestamps
                                          // event variable: number of timestamps
                                          // interval variable: number of intervals
                                          // waveform variable: number of waveforms
                                          // continuous variable: number of fragments
                                          // population vector: number of weights
        public int TimestampDataType { get; set; } // if 0, timestamps are stored as 32-bit integers; 
                                                   // if 1, timestamps are stored as 64-bit integers;  not supported in NeuroExplorer yet (version 5.016)
        public int ContinuousDataType { get; set; } // waveforms and continuous variables only, 
                                                    // if 0, waveform and continuous values are stored as 16-bit integers; 
                                                    // if 1, waveform and continuous values are stored as 32-bit floating point values in units specified in Units field
        public double SamplingFrequency { get; set; } // waveforms and continuous variables only, waveform or continuous variable sampling frequency in Hertz
        public string Units { get; set; } // waveforms and continuous variables only, units that should be used for the variable values
                                          // not supported as of July 2015 (version 5.016)
        public double ADtoUnitsCoefficient { get; set; } // waveforms and continuous variables only, coefficient to convert from A/D values to units.
                                                         // see formula below UnitsOffset below; ignored if ContinuousDataType == 1 
        public double UnitsOffset { get; set; } // waveforms and continuous variables only, 
                                                // this offset is used to convert A/D values in units:
                                                //  value_in_units = raw * ADtoUnitsCoefficient + UnitsOffset; ignored if ContinuousDataType == 1 
        public ulong NumberOfDataPoints { get; set; } // waveform variable: number of data points in each wave
                                                       // continuous variable: overall number of data points in the variable
        public double PrethresholdTimeInSeconds { get; set; } // waveform variables only, pre-threshold time in seconds
                                                              // if waveform timestamp in seconds is t, 
                                                              // then the timestamp of the first point of waveform is t - PrethresholdTimeInSeconds
        public int MarkerDataType { get; set; } // marker events only,
                                                // if 0, marker values are stored as strings; 
                                                // if 1, marker values are stored as 32-bit unsigned integers
        public int NumberOfMarkerFields { get; set; } // marker events only, how many values are associated with each marker
        public int MarkerLength { get; set; } // marker events only, how many characters are in each marker value; ignored if MarkerDataType is 1
        public int ContinuousIndexOfFirstPointInFragmentDataType { get; set; } //continuous variables only, 
                                                                               // if 0, indexes of first data point in fragments are stored as unsigned 32-bit integers; 
                                                                               // if 1, indexes of first data point in fragments are stored as unsigned 64-bit integers; not supported in NeuroExplorer yet (version 5.016)

        /// <summary>
        /// Initializes a new instance of the <see cref="VariableHeader"/> class.
        /// </summary>
        public Nex5FileVarHeader()
        {
            Type = VariableType.Neuron;
            Version = 500;
            Name = "(no name)";
            DataOffset = 0;
            Count = 0;
            TimestampDataType = 0;
            ContinuousDataType = 0;
            SamplingFrequency = 0;
            Units = "mV";
            ADtoUnitsCoefficient = 1.0;
            UnitsOffset = 0.0;
            NumberOfDataPoints = 0;
            PrethresholdTimeInSeconds = 0;
            MarkerDataType = 0;
            NumberOfMarkerFields = 0;
            MarkerLength = 0;
            ContinuousIndexOfFirstPointInFragmentDataType = 0;
        }


        public void ReadFromFile( BinaryReader reader )
        {
            Type = ( VariableType )reader.ReadInt32();
            Version = reader.ReadInt32();
            Name = Encoding.UTF8.GetString( reader.ReadBytes( 64 ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() );
            DataOffset = reader.ReadUInt64();
            Count = reader.ReadUInt64();
            TimestampDataType = reader.ReadInt32();
            ContinuousDataType = reader.ReadInt32();
            SamplingFrequency = reader.ReadDouble();
            Units = Encoding.UTF8.GetString( reader.ReadBytes( 32 ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() );
            ADtoUnitsCoefficient = reader.ReadDouble();
            UnitsOffset = reader.ReadDouble();
            NumberOfDataPoints = reader.ReadUInt64();
            PrethresholdTimeInSeconds = reader.ReadDouble();
            MarkerDataType = reader.ReadInt32();
            NumberOfMarkerFields = reader.ReadInt32();
            MarkerLength = reader.ReadInt32();
            ContinuousIndexOfFirstPointInFragmentDataType = reader.ReadInt32();
            var padding = reader.ReadBytes( 60 );
        }

        public void WriteToFile( BinaryWriter writer )
        {
            writer.Write( ( int )Type );
            writer.Write( Version );
            writer.Write( Utilities.GetBytesFromStringUTF8( Name, 64 ) );
            writer.Write( DataOffset );
            writer.Write( Count );
            writer.Write( TimestampDataType );
            writer.Write( ContinuousDataType );
            writer.Write( SamplingFrequency );
            writer.Write( Utilities.GetBytesFromStringUTF8( Units, 32 ) );
            writer.Write( ADtoUnitsCoefficient );
            writer.Write( UnitsOffset );
            writer.Write( NumberOfDataPoints );
            writer.Write( PrethresholdTimeInSeconds );
            writer.Write( MarkerDataType );
            writer.Write( NumberOfMarkerFields );
            writer.Write( MarkerLength );
            writer.Write( ContinuousIndexOfFirstPointInFragmentDataType );
            writer.Write( new byte[ 60 ] );
        }
    }
}
