﻿using System;
using System.Collections.Generic;
using System.IO;
using Newtonsoft.Json;

namespace NeuroExplorer.NexFile
{
    public class Nex5FileWriter
    {
        private const int BytesInNex5FileHeader = 356;
        private const int BytesInNex5VariableHeader = 244;


        public void WriteDataToNex5File( FileData fd, string filePath )
        {
            using ( FileStream stream = new FileStream( filePath, FileMode.OpenOrCreate, FileAccess.Write, FileShare.None ) ) {
                using ( BinaryWriter writer = new BinaryWriter( stream ) ) {
                    WriteNex5FileHeader( fd, writer );
                    WriteNex5VariableHeaders( fd, writer );
                    WriteNex5VariableData( fd, writer );
                    WriteMetaData( fd, writer );
                    writer.Close();
                }
                stream.Close();
            }
        }


        private void WriteNex5VariableData( FileData fd, BinaryWriter writer )
        {
            foreach ( var neuron in fd.NeuronVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToLongTicks( neuron.Timestamps ) );
            }
            foreach ( var nexEvent in fd.EventVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToLongTicks( nexEvent.Timestamps ) );
            }
            foreach ( var interval in fd.IntervalVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToLongTicks( interval.IntervalStarts ) );
                Utilities.WriteListOfNumbers( writer, fd.SecondsToLongTicks( interval.IntervalEnds ) );
            }
            foreach ( var marker in fd.MarkerVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToLongTicks( marker.Timestamps ) );
                for ( int i = 0; i < marker.MarkerNames.Count; i++ ) {
                    writer.Write( Utilities.GetBytesFromString( marker.MarkerNames[ i ], 64 ) );
                    if ( marker.MarkerValues.Count == 0 ) {
                        // writing strings
                        foreach ( var mk in marker.MarkerStrings[ i ] ) {
                            writer.Write( Utilities.GetBytesFromString( mk, marker.MaxMarkerLength ) );
                        }
                    } else {
                        // writing numbers
                        Utilities.WriteListOfNumbers( writer, marker.MarkerValues[ i ] );
                    }
                }
            }

            foreach ( var cont in fd.ContinuousVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToLongTicks( cont.FragmentStartsTimestamps ) );
                Utilities.WriteListOfNumbers( writer, cont.FragmentStartsIndexes );
                Utilities.WriteListOfNumbers( writer, cont.Values );
            }

            foreach ( var wave in fd.WaveformVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToLongTicks( wave.Timestamps ) );
                foreach ( var w in wave.Values ) {
                    Utilities.WriteListOfNumbers( writer, w );
                }
            }
        }

        private void WriteNex5VariableHeaders( FileData fd, BinaryWriter writer )
        {
            ulong dataOffset = ( ulong )BytesInNex5FileHeader + ( ulong )fd.NumberOfVariables() * ( ulong )BytesInNex5VariableHeader;

            foreach ( var neuron in fd.NeuronVariables ) {
                Nex5FileVarHeader header = new Nex5FileVarHeader();
                header.Type = VariableType.Neuron;
                header.Name = neuron.Name;
                header.Count = ( ulong )neuron.Timestamps.Count;
                header.TimestampDataType = 1;
                header.DataOffset = dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 8;
            }

            foreach ( var nexEvent in fd.EventVariables ) {
                Nex5FileVarHeader header = new Nex5FileVarHeader();
                header.Type = VariableType.Event;
                header.Name = nexEvent.Name;
                header.Count = ( ulong )nexEvent.Timestamps.Count;
                header.TimestampDataType = 1;
                header.DataOffset = dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 8;
            }

            foreach ( var interval in fd.IntervalVariables ) {
                Nex5FileVarHeader header = new Nex5FileVarHeader();
                header.Type = VariableType.Interval;
                header.Name = interval.Name;
                header.Count = ( ulong )interval.IntervalStarts.Count;
                header.TimestampDataType = 1;
                header.DataOffset = dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 8 * 2;
            }

            foreach ( var marker in fd.MarkerVariables ) {
                Nex5FileVarHeader header = new Nex5FileVarHeader();
                header.Type = VariableType.Marker;
                header.Name = marker.Name;
                header.Count = ( ulong )marker.Timestamps.Count;
                header.TimestampDataType = 1;
                header.NumberOfMarkerFields = marker.MarkerNames.Count;
                header.MarkerLength = 0;
                header.MarkerDataType = 1;
                int bytesInMarker = 4;
                if ( marker.MarkerValues.Count == 0 ) {
                    bytesInMarker = 0;
                    foreach ( var field in marker.MarkerStrings ) {
                        foreach ( var mk in field ) {
                            bytesInMarker = Math.Max( bytesInMarker, mk.Length );
                        }
                    }
                    header.MarkerLength = bytesInMarker;
                    marker.MaxMarkerLength = bytesInMarker;
                    header.MarkerDataType = 0;
                }

                header.DataOffset = dataOffset;
                header.WriteToFile( writer );

                dataOffset += ( ulong )header.Count * 8
                    + ( ulong )header.NumberOfMarkerFields * 64
                    + ( ulong )header.NumberOfMarkerFields * ( ulong )bytesInMarker * header.Count;
            }

            foreach ( var cont in fd.ContinuousVariables ) {
                Nex5FileVarHeader header = new Nex5FileVarHeader();
                header.Type = VariableType.Continuous;
                header.Name = cont.Name;
                header.SamplingFrequency = cont.SamplingRate;
                header.Count = ( ulong )cont.FragmentStartsTimestamps.Count;
                header.NumberOfDataPoints = ( ulong )cont.Values.Count;
                header.TimestampDataType = 1;
                header.ContinuousDataType = 1;
                header.ContinuousIndexOfFirstPointInFragmentDataType = 0;
                header.DataOffset = dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * ( 8 + 4 ) + header.NumberOfDataPoints * 4;
            }

            foreach ( var wave in fd.WaveformVariables ) {
                Nex5FileVarHeader header = new Nex5FileVarHeader();
                header.Type = VariableType.Waveform;
                header.Name = wave.Name;
                header.SamplingFrequency = wave.SamplingRate;
                header.Count = ( ulong )wave.Timestamps.Count;
                header.TimestampDataType = 1;
                header.NumberOfDataPoints = ( ulong )wave.NPointsWave;
                header.ContinuousDataType = 1;
                header.PrethresholdTimeInSeconds = wave.PrethresholdTimeInSeconds;

                header.DataOffset = dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 8 + header.Count * header.NumberOfDataPoints * 4;
            }
        }

        private void WriteNex5FileHeader( FileData fd, BinaryWriter writer )
        {
            Nex5FileHeader fh = new Nex5FileHeader();
            fh.Nex5MagicNumber = 0x3558454e;
            fh.Nex5FileVersion = 502;
            fh.Comment = fd.Comment;
            fh.TimestampFrequency = fd.TimestampFrequency;
            fh.RecordingStartTimeInTicks = fd.SecondsToLongTicks( fd.BegSeconds );
            fh.NumberOfVariables = fd.NumberOfVariables();
            fh.MetadataOffset = 0;
            fh.RecordingEndTimeInTicks = fd.SecondsToLongTicks( fd.MaxTimestamp() );
            fh.WriteToFile( writer );
        }

        private void WriteMetaData( FileData fd, BinaryWriter writer )
        {
            Nex5Meta meta = new Nex5Meta();

            meta.file[ "WriterSoftwareName" ] = "NeuroExplorer.NexFile";

            System.Reflection.Assembly assembly = System.Reflection.Assembly.GetExecutingAssembly();
            System.Diagnostics.FileVersionInfo fvi = System.Diagnostics.FileVersionInfo.GetVersionInfo( assembly.Location );
            string version = fvi.FileVersion;
            meta.file[ "WriterSoftware Version" ] = version;

            foreach ( var item in fd.ExtraFileMetaData ) {
                meta.file[ item.Key ] = item.Value;
            }

            foreach ( var neuron in fd.NeuronVariables ) {
                Nex5VarMeta vm = new Nex5VarMeta();
                vm.name = neuron.Name;
                vm.unitNumber = neuron.UnitNumber;
                vm.probe = new Probe();
                vm.probe.wireNumber = neuron.WireNumber;
                vm.probe.position = new Position();
                vm.probe.position.x = neuron.XPos;
                vm.probe.position.y = neuron.YPos;
                meta.variables.Add( vm );
            }

            foreach ( var wave in fd.WaveformVariables ) {
                Nex5VarMeta vm = new Nex5VarMeta();
                vm.name = wave.Name;
                vm.unitNumber = wave.UnitNumber;
                vm.probe = new Probe();
                vm.probe.wireNumber = wave.WireNumber;
                meta.variables.Add( vm );
            }

            string json = JsonConvert.SerializeObject( meta );
            long pos = writer.BaseStream.Position;
            writer.Write( System.Text.Encoding.UTF8.GetBytes( json ) );
            writer.Seek( 284, SeekOrigin.Begin );
            writer.Write( pos );
        }
    }
}
