﻿using System;
using System.IO;
using System.Linq;
using System.Text;

//// .nex file header structure
//struct NexFileHeader
//{
//    int MagicNumber; // string NEX1; numeric value: 827868494 decimal or 0x3158454E hex
//    int NexFileVersion; // current version (Feb-2011) is 106;
//    // when reading existing .nex files, use the following rules:
//    // Versions 102 and 103: beta versions; should not be used
//    // Versions 100, 101 and 104: use standard .nex data read algorithm as described,
//    //     except that NexVarHeader::MVOffset and NexVarHeader::PrethresholdTimeInSeconds are not used and are always zero
//    // Version 105: indicates that NexVarHeader::MVOffset can be non-zero
//    // Version 106: indicates that NexVarHeader::PrethresholdTimeInSeconds can be non-zero

//    char Comment[ 256 ]; // file comment
//    double Frequency;  // timestamps frequency, in Hertz
//                       // timestamp values are stored in ticks, where tick = 1/Frequency.
//    int Beg;    // minimum timestamp
//    int End;    // maximum timestamp + 1
//    int NumVars;    // number of variables in the file
//    int NextFileHeader; // position of the next file header in the file
//                        // this field is not implemented yet
//    char Padding[ 256 ]; // padding for future expansion
//};

namespace NeuroExplorer.NexFile
{
    /// <summary>
    /// File header in .nex file
    /// </summary>
    public class NexFileHeader
    {
        public int MagicNumber { get; set; } // string NEX1; numeric value: 827868494 decimal or 0x3158454E hex
        public int NexFileVersion { get; set; }// current version (Feb-2011) is 106;
                                               // when reading existing .nex files, use the following rules:
                                               // Versions 102 and 103: beta versions; should not be used
                                               // Versions 100, 101 and 104: use standard .nex data read algorithm as described,
                                               //     except that NexVarHeader::MVOffset and NexVarHeader::PrethresholdTimeInSeconds are not used and are always zero
                                               // Version 105: indicates that NexVarHeader::MVOffset can be non-zero
                                               // Version 106: indicates that NexVarHeader::PrethresholdTimeInSeconds can be non-zero

        public string Comment { get; set; } // file comment
        public double TimestampFrequency { get; set; }  // timestamps frequency, in Hertz
                                               // timestamp values are stored in ticks, where tick = 1/Frequency.
        public int Beg { get; set; } // minimum timestamp in ticks
        public int End { get; set; }    // maximum timestamp in ticks
        public int NumVars { get; set; }  // number of variables in the file

        public NexFileHeader()
        {
            MagicNumber = 0;
            NexFileVersion = 0;
            Comment = "";
            TimestampFrequency = 0;
            Beg = 0;
            End = 0;
            NumVars = 0;
        }

        public void ReadFromFile( BinaryReader reader )
        {
            MagicNumber = reader.ReadInt32();
            if ( MagicNumber != 0x3158454e ) {
                throw new ApplicationException( "Invalid .nex file: magic number in the file header is incorrect" );
            }
            NexFileVersion = reader.ReadInt32();
            Comment = Encoding.ASCII.GetString( reader.ReadBytes( 256 ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() );
            TimestampFrequency = reader.ReadDouble();
            if ( TimestampFrequency <= 0 ) {
                throw new ApplicationException( "Invalid .nex file: timestamp frequency is not positive" );
            }
            Beg = reader.ReadInt32();
            End = reader.ReadInt32();
            NumVars = reader.ReadInt32();
            var nextHeader = reader.ReadInt32();
            var padding = reader.ReadBytes( 256 );
        }

        public void WriteToFile( BinaryWriter writer )
        {
            writer.Write(MagicNumber);
            writer.Write( NexFileVersion );
            writer.Write( Utilities.GetBytesFromString( Comment, 256 ) );
            writer.Write( TimestampFrequency );
            writer.Write( Beg );
            writer.Write( End );
            writer.Write( NumVars );
            writer.Write( new byte[ 260 ] );
        }
    }
}
