﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace NeuroExplorer.NexFile
{
    public class NexFileReader
    {
        public FileData ReadNexFile( string filePath )
        {
            FileData fd = new FileData();
            using ( FileStream stream = new FileStream( filePath, FileMode.Open, FileAccess.Read, FileShare.Read ) ) {
                using ( BinaryReader reader = new BinaryReader( stream ) ) {
                    
                    // 1. read file header
                    NexFileHeader fileHeader = new NexFileHeader();
                    fileHeader.ReadFromFile( reader );

                    fd.TimestampFrequency = fileHeader.TimestampFrequency;
                    fd.Comment = fileHeader.Comment;
                    fd.BegSeconds = fileHeader.Beg / fileHeader.TimestampFrequency;
                    fd.EndSeconds = fileHeader.End / fileHeader.TimestampFrequency;

                    // 2. read variable headers
                    List<NexFileVarHeader> nexVarHeaders = new List<NexFileVarHeader>();
                    for ( int i = 0; i < fileHeader.NumVars; i++ ) {
                        NexFileVarHeader varHeader = new NexFileVarHeader();
                        varHeader.ReadFromFile( reader );
                        nexVarHeaders.Add( varHeader );
                    }

                    // 3. read data
                    foreach ( var vh in nexVarHeaders ) {
                        switch ( vh.Type ) {
                            case VariableType.Event:
                                Event nexEvent = new Event();
                                nexEvent.Name = vh.Name;
                                reader.BaseStream.Seek( vh.DataOffset, SeekOrigin.Begin );
                                nexEvent.Timestamps = Utilities.ReadIntTimestamps( reader, vh.Count, fileHeader.TimestampFrequency );
                                fd.EventVariables.Add( nexEvent );
                                break;

                            case VariableType.Neuron:
                                Neuron neuron = new Neuron();
                                neuron.Name = vh.Name;
                                neuron.WireNumber = vh.WireNumber;
                                neuron.UnitNumber = vh.UnitNumber;
                                neuron.XPos = vh.XPos;
                                neuron.YPos = vh.YPos;
                                reader.BaseStream.Seek( vh.DataOffset, SeekOrigin.Begin );
                                neuron.Timestamps = Utilities.ReadIntTimestamps( reader, vh.Count, fileHeader.TimestampFrequency );
                                fd.NeuronVariables.Add( neuron );
                                break;

                            case VariableType.Interval:
                                Interval interval = new Interval();
                                interval.Name = vh.Name;
                                reader.BaseStream.Seek( vh.DataOffset, SeekOrigin.Begin );
                                interval.IntervalStarts = Utilities.ReadIntTimestamps( reader, vh.Count, fileHeader.TimestampFrequency );
                                interval.IntervalEnds = Utilities.ReadIntTimestamps( reader, vh.Count, fileHeader.TimestampFrequency );
                                fd.IntervalVariables.Add( interval );
                                break;

                            case VariableType.Marker:
                                Marker marker = new Marker();
                                marker.Name = vh.Name;
                                reader.BaseStream.Seek( vh.DataOffset, SeekOrigin.Begin );
                                marker.Timestamps = Utilities.ReadIntTimestamps( reader, vh.Count, fileHeader.TimestampFrequency );
                                for ( int i = 0; i < vh.NMarkers; i++ ) {
                                    marker.MarkerNames.Add( Encoding.ASCII.GetString( reader.ReadBytes( 64 ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() ) );
                                    var s = new List<string>();
                                    for ( int j = 0; j < vh.Count; j++ ) {
                                        s.Add( Encoding.ASCII.GetString( reader.ReadBytes( vh.MarkerLength ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() ) );
                                    }
                                    marker.MarkerStrings.Add( s );
                                }
                                fd.MarkerVariables.Add( marker );
                                break;

                            case VariableType.Continuous:
                                Continuous cont = new Continuous();
                                cont.Name = vh.Name;
                                cont.SamplingRate = vh.WFrequency;
                                reader.BaseStream.Seek( vh.DataOffset, SeekOrigin.Begin );
                                cont.FragmentStartsTimestamps = Utilities.ReadIntTimestamps( reader, vh.Count, fileHeader.TimestampFrequency );
                                cont.FragmentStartsIndexes = Utilities.ReadUIntsArrayAsList( reader, vh.Count );
                                cont.Values = Utilities.ReadAndScaleShortsArray( reader, vh.NPointsWave, vh.ADtoMV, vh.MVOffset );
                                fd.ContinuousVariables.Add( cont );
                                break;

                            case VariableType.Waveform:
                                Waveform wave = new Waveform();
                                wave.Name = vh.Name;
                                wave.SamplingRate = vh.WFrequency;
                                wave.NPointsWave = vh.NPointsWave;
                                reader.BaseStream.Seek( vh.DataOffset, SeekOrigin.Begin );
                                wave.Timestamps = Utilities.ReadIntTimestamps( reader, vh.Count, fileHeader.TimestampFrequency );
                                wave.Values = Utilities.ReadAndScaleShortsArrayWaves( reader, vh.Count, vh.ADtoMV, vh.MVOffset, vh.NPointsWave );
                                fd.WaveformVariables.Add( wave );
                                break;
                        }
                    }
                }
            }
            return fd;
        }
    }
}
