﻿using System.IO;
using System.Linq;
using System.Text;

// .nex file variable header structure
//struct NexVarHeader
//{
//    int Type; // 0 - neuron, 1 - event, 2- interval, 3 - waveform, 4 - pop. vector, 5 - continuously recorded, 6 - marker
//              // see #define statements above
//    int Version; // almost always should be Version = 100; 
//                 // use Version = 101 if it is a neuron or waveform variable and WireNumber and UnitNumber are valid 
//                 // use Version = 102 if it is a waveform variable and PrethresholdTimeInSeconds is valid
//    char Name[ 64 ]; // variable name
//    int DataOffset; // where the data array for this variable is located in the file
//    int Count; // neuron variable: number of timestamps
//               // event variable: number of timestamps
//               // interval variable: number of intervals
//               // waveform variable: number of waveforms
//               // continuous variable: number of fragments
//               // population vector: number of weights
//    int WireNumber; // neurons and waveforms only; for data from PLX files, channel number from the record header
//    int UnitNumber; // neurons and waveforms only; for data from PLX files, unit number from the record header
//    int Gain; // neurons only, not used
//    int Filter; // neurons only, not used
//    double XPos; // neurons only, X axis electrode position in (0,100) range, used in 3D display
//    double YPos; //  neurons only, Y axis electrode position in (0,100) range, used in 3D display
//    double WFrequency; // waveforms and continuous variables only, w/f or cont. sampling frequency in Hertz
//    double ADtoMV; // waveforms and continuous variables only, coeff. to convert from A/D values to Millivolts.
//                   // see formula below MVOffset below
//    int NPointsWave; // waveform variable: number of points in each wave
//                     // continuous variable: number of data points
//    int NMarkers; // marker events only, how many values are associated with each marker
//    int MarkerLength; // marker events only, how many characters are in each marker value
//    double MVOffset; // waveforms and continuous variables only, 
//                     // this offset is used to convert A/D values in Millivolts:
//                     //  mv = raw * ADtoMV + MVOffset
//    double PrethresholdTimeInSeconds; // for waveforms, pre-threshold time in seconds
//                                      // if waveform timestamp in seconds is t, 
//                                      // then the timestamp of the first point of waveform is t - PrethresholdTimeInSeconds
//    char Padding[ 52 ]; // padding for future expansion
//};

namespace NeuroExplorer.NexFile
{
    public class NexFileVarHeader
    {
        public VariableType Type { get; set; } // 0 - neuron, 1 - event, 2- interval, 3 - waveform, 4 - pop. vector, 5 - continuously recorded, 6 - marker
                                               // see #define statements above
        public int Version { get; set; } // almost always should be Version = 100
                                         // use Version = 101 if it is a neuron or waveform variable and WireNumber and UnitNumber are valid 
                                         // use Version = 102 if it is a waveform variable and PrethresholdTimeInSeconds is valid
        public string Name { get; set; } // variable name
        public int DataOffset { get; set; } // where the data array for this variable is located in the file
        public int Count { get; set; } // neuron variable: number of timestamps
                                       // event variable: number of timestamps
                                       // interval variable: number of intervals
                                       // waveform variable: number of waveforms
                                       // continuous variable: number of fragments
                                       // population vector: number of weights
        public int WireNumber { get; set; } // neurons and waveforms only. channel number from the record header
        public int UnitNumber { get; set; } // neurons and waveforms only. unit number from the record header
        public double XPos { get; set; } // neurons only, X axis electrode position in (0,100) range, used in 3D display
        public double YPos { get; set; } //  neurons only, Y axis electrode position in (0,100) range, used in 3D display
        public double WFrequency { get; set; } // waveforms and continuous variables only, w/f or cont. sampling frequency in Hertz
        public double ADtoMV { get; set; } // waveforms and continuous variables only, coeff. to convert from A/D values to Millivolts.
                                           // see formula below MVOffset below
        public int NPointsWave { get; set; } // waveform variable: number of points in each wave
                                             // continuous variable: number of data points
        public int NMarkers { get; set; } // marker events only, how many values are associated with each marker
        public int MarkerLength { get; set; } // marker events only, how many characters are in each marker value
        public double MVOffset { get; set; } // waveforms and continuous variables only, 
                                             // this offset is used to convert A/D values in Millivolts:
                                             //  mv = raw * ADtoMV + MVOffset
        public double PrethresholdTimeInSeconds { get; set; } // for waveforms, pre-threshold time in seconds
                                                              // if waveform timestamp in seconds is t, 
                                                              // then the timestamp of the first point of waveform is t - PrethresholdTimeInSeconds

        public NexFileVarHeader()
        {
            Type = VariableType.Neuron;
            Version = 100;
            Name = "(no name)";
            DataOffset = 0;
            Count = 0;
            WireNumber = 0;
            UnitNumber = 0;
            XPos = 0;
            YPos = 0;
            WFrequency = 0;
            ADtoMV = 0;
            NPointsWave = 0;
            NMarkers = 0;
            MarkerLength = 0;
            MVOffset = 0;
            PrethresholdTimeInSeconds = 0;
        }

        public void ReadFromFile( BinaryReader reader )
        {
            Type = ( VariableType )reader.ReadInt32();
            Version = reader.ReadInt32();
            Name = Encoding.ASCII.GetString( reader.ReadBytes( 64 ).TakeWhile( b => !b.Equals( 0 ) ).ToArray() );
            DataOffset = reader.ReadInt32();
            Count = reader.ReadInt32();
            WireNumber = reader.ReadInt32();
            UnitNumber = reader.ReadInt32();
            var gain = reader.ReadInt32();
            var filter = reader.ReadInt32();
            XPos = reader.ReadDouble();
            YPos = reader.ReadDouble();
            WFrequency = reader.ReadDouble();
            ADtoMV = reader.ReadDouble();
            NPointsWave = reader.ReadInt32();
            NMarkers = reader.ReadInt32();
            MarkerLength = reader.ReadInt32();
            MVOffset = reader.ReadDouble();
            PrethresholdTimeInSeconds = reader.ReadDouble();
            var padding = reader.ReadBytes( 52 );
        }

        public void WriteToFile( BinaryWriter writer )
        {
            writer.Write( ( int )Type );
            writer.Write( Version );
            writer.Write( Utilities.GetBytesFromString( Name, 64 ) );
            writer.Write( DataOffset );
            writer.Write( Count );
            writer.Write( WireNumber );
            writer.Write( UnitNumber );
            writer.Write( new byte[ 8 ] );
            writer.Write( XPos );
            writer.Write( YPos );
            writer.Write( WFrequency );
            writer.Write( ADtoMV );
            writer.Write( NPointsWave );
            writer.Write( NMarkers );
            writer.Write( MarkerLength );
            writer.Write( MVOffset );
            writer.Write( PrethresholdTimeInSeconds );
            writer.Write( new byte[ 52 ] );
        }

    }
}
