﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace NeuroExplorer.NexFile
{
    public class NexFileWriter
    {
        private const int BytesInNexFileHeader = 544;
        private const int BytesInNexVariableHeader = 208;


        public void WriteDataToNexFile( FileData fd, string filePath )
        {
            using ( FileStream stream = new FileStream( filePath, FileMode.OpenOrCreate, FileAccess.Write, FileShare.None ) ) {
                using ( BinaryWriter writer = new BinaryWriter( stream ) ) {
                    WriteNexFileHeader( fd, writer );
                    WriteNexVariableHeaders( fd, writer );
                    WriteNexVariableData( fd, writer );
                    writer.Close();
                }
                stream.Close();
            }
        }

        private void WriteNexVariableData( FileData fd, BinaryWriter writer )
        {
            foreach ( var neuron in fd.NeuronVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToIntTicks( neuron.Timestamps ) );
            }
            foreach ( var nexEvent in fd.EventVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToIntTicks( nexEvent.Timestamps ) );
            }
            foreach ( var interval in fd.IntervalVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToIntTicks( interval.IntervalStarts ) );
                Utilities.WriteListOfNumbers( writer, fd.SecondsToIntTicks( interval.IntervalEnds ) );
            }
            foreach ( var marker in fd.MarkerVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToIntTicks( marker.Timestamps ) );
                for ( int i = 0; i < marker.MarkerNames.Count; i++ ) {
                    writer.Write( Utilities.GetBytesFromString( marker.MarkerNames[ i ], 64 ) );
                    foreach ( var mk in marker.MarkerStrings[ i ] ) {
                        writer.Write( Utilities.GetBytesFromString( mk, marker.MaxMarkerLength ) );
                    }
                }
            }

            foreach ( var cont in fd.ContinuousVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToIntTicks( cont.FragmentStartsTimestamps ) );
                Utilities.WriteListOfNumbers( writer, cont.FragmentStartsIndexes );
                Utilities.WriteListOfNumbers( writer, Utilities.ConvertContValuesToShorts( cont ) );
            }

            foreach ( var wave in fd.WaveformVariables ) {
                Utilities.WriteListOfNumbers( writer, fd.SecondsToIntTicks( wave.Timestamps ) );
                Utilities.WriteListOfNumbers( writer, Utilities.ConvertWaveValuesToShorts( wave ) );
            }
        }

        private void WriteNexVariableHeaders( FileData fd, BinaryWriter writer )
        {
            long dataOffset = BytesInNexFileHeader + fd.NumberOfVariables() * BytesInNexVariableHeader;

            foreach ( var neuron in fd.NeuronVariables ) {
                NexFileVarHeader header = new NexFileVarHeader();
                header.Type = VariableType.Neuron;
                header.Name = neuron.Name;
                header.Count = neuron.Timestamps.Count;
                header.WireNumber = neuron.WireNumber;
                header.UnitNumber = neuron.UnitNumber;
                header.XPos = neuron.XPos;
                header.YPos = neuron.YPos;
                header.DataOffset = ( int )dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 4;
            }

            foreach ( var nexEvent in fd.EventVariables ) {
                NexFileVarHeader header = new NexFileVarHeader();
                header.Type = VariableType.Event;
                header.Name = nexEvent.Name;
                header.Count = nexEvent.Timestamps.Count;
                header.DataOffset = ( int )dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 4;
            }

            foreach ( var interval in fd.IntervalVariables ) {
                NexFileVarHeader header = new NexFileVarHeader();
                header.Type = VariableType.Interval;
                header.Name = interval.Name;
                header.Count = interval.IntervalStarts.Count;
                header.DataOffset = ( int )dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 8;
            }

            foreach ( var marker in fd.MarkerVariables ) {
                NexFileVarHeader header = new NexFileVarHeader();
                header.Type = VariableType.Marker;
                header.Name = marker.Name;
                header.Count = marker.Timestamps.Count;
                header.NMarkers = marker.MarkerNames.Count;
                header.MarkerLength = 0;
                int bytesInMarker = 0;
                foreach ( var field in marker.MarkerStrings ) {
                    foreach ( var mk in field ) {
                        bytesInMarker = Math.Max( bytesInMarker, mk.Length );
                    }
                }
                header.MarkerLength = bytesInMarker;
                marker.MaxMarkerLength = bytesInMarker;

                header.DataOffset = ( int )dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 4
                    + header.NMarkers * 64
                    + header.NMarkers * bytesInMarker * header.Count;
            }

            foreach ( var cont in fd.ContinuousVariables ) {
                NexFileVarHeader header = new NexFileVarHeader();
                header.Type = VariableType.Continuous;
                header.Name = cont.Name;
                header.WFrequency = cont.SamplingRate;
                header.Count = cont.FragmentStartsTimestamps.Count;
                header.NPointsWave = cont.Values.Count;
                header.ADtoMV = 1.0 / Utilities.CalcScaleFloatsToShorts( cont.Values );
                cont.CalulatedCoeffToScaleFloatsToShorts = 1.0 / header.ADtoMV;
                header.MVOffset = 0;
                header.DataOffset = ( int )dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * ( 4 + 4 ) + header.NPointsWave * 2;
            }

            foreach ( var wave in fd.WaveformVariables ) {
                NexFileVarHeader header = new NexFileVarHeader();
                header.Type = VariableType.Waveform;
                header.Name = wave.Name;
                header.WFrequency = wave.SamplingRate;
                header.Count = wave.Timestamps.Count;
                header.NPointsWave = wave.NPointsWave;
                header.ADtoMV = 1.0 / Utilities.CalcScaleFloatsToShorts( wave.Values );
                wave.CalulatedCoeffToScaleFloatsToShorts = 1.0 / header.ADtoMV;
                header.MVOffset = 0;
                header.PrethresholdTimeInSeconds = wave.PrethresholdTimeInSeconds;

                header.DataOffset = ( int )dataOffset;
                header.WriteToFile( writer );

                dataOffset += header.Count * 4 + header.Count * header.NPointsWave * 2;
            }

            if ( dataOffset > 0xFFFFFFFF ) {
                throw new ApplicationException( "Unable to save data in .nex file: file size exceeds 2^32" );
            }
        }

        private void WriteNexFileHeader( FileData fd, BinaryWriter writer )
        {
            NexFileHeader fh = new NexFileHeader();
            fh.MagicNumber = 0x3158454e;
            fh.NexFileVersion = 106;
            fh.Comment = fd.Comment;
            fh.TimestampFrequency = fd.TimestampFrequency;
            fh.Beg = fd.SecondsToIntTicks( fd.BegSeconds );
            long maxTick = fd.SecondsToLongTicks( fd.MaxTimestamp() );
            if ( maxTick > 0x7FFFFFFF ) {
                throw new ApplicationException( "Unable to save data in .nex file: maximum timestamp exceeds 2^31" );
            }
            fh.End = ( int )maxTick;
            fh.NumVars = fd.NumberOfVariables();
            fh.WriteToFile( writer );
        }
    }
}
