﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace NeuroExplorer.NexFile
{
    /// <summary>
    /// Types of variables in .nex and .nex5 data files.
    /// </summary>
    public enum VariableType
    {
        /// <summary>
        /// Neuron variable type. See <see cref="Neuron"/> class.
        /// </summary>
        Neuron = 0,
        /// <summary>
        /// Event variable type. See <see cref="Event"/> class.
        /// </summary>
        Event,
        /// <summary>
        /// Interval variable type. See <see cref="Interval"/> class.
        /// </summary>
        Interval,
        /// <summary>
        /// Waveform variable type. See <see cref="Waveform"/> class.
        /// </summary>
        Waveform,
        /// <summary>
        /// Population vector variable type. Not implemented in this library.
        /// </summary>
        PopulationVector,
        /// <summary>
        /// Continuous variable type. See <see cref="Continuous"/> class.
        /// </summary>
        Continuous,
        /// <summary>
        /// Marker variable type. See <see cref="Marker"/> class.
        /// </summary>
        Marker
    }

    public static class Utilities
    {
        public static List<int> ReadIntsArrayAsList( BinaryReader reader, int count )
        {
            var byteArray = reader.ReadBytes( count * 4 );
            int[] values = new int[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            return new List<int>( values );
        }

        public static List<uint> ReadUIntsArrayAsList( BinaryReader reader, int count )
        {
            var byteArray = reader.ReadBytes( count * 4 );
            uint[] values = new uint[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            return new List<uint>( values );
        }


        public static List<int> ReadUIntsArrayAsListReturnUIntList( BinaryReader reader, int count )
        {
            var byteArray = reader.ReadBytes( count * 4 );
            uint[] values = new uint[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            var theList = new List<int>();
            foreach ( var x in values ) {
                theList.Add( ( int )x );
            }
            return theList;
        }

        public static List<uint> ReadUInts64ArrayAsListReturnUIntList( BinaryReader reader, int count )
        {
            var byteArray = reader.ReadBytes( count * 8 );
            ulong[] values = new ulong[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            var theList = new List<uint>();
            foreach ( var x in values ) {
                theList.Add( ( uint )x );
            }
            return theList;
        }

        public static List<double> ReadIntTimestamps( BinaryReader reader, int count, double timestampFrequency )
        {
            var byteArray = reader.ReadBytes( count * 4 );
            int[] values = new int[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            var ts = new List<double>();
            foreach ( var t in values ) {
                ts.Add( t / timestampFrequency );
            }
            return ts;
        }

        public static List<double> ReadInt64Timestamps( BinaryReader reader, int count, double timestampFrequency )
        {
            var byteArray = reader.ReadBytes( count * 8 );
            long[] values = new long[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            var ts = new List<double>();
            foreach ( var t in values ) {
                ts.Add( t / timestampFrequency );
            }
            return ts;
        }

        public static List<float> ReadFloatsArray( BinaryReader reader, int count )
        {
            var byteArray = reader.ReadBytes( count * 4 );
            float[] values = new float[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            return new List<float>( values );
        }

        public static List<float> ReadAndScaleShortsArray( BinaryReader reader, int count, double coeff, double offset )
        {
            var byteArray = reader.ReadBytes( count * 2 );
            short[] values = new short[ count ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            var floats = new List<float>();
            foreach ( var v in values ) {
                floats.Add( ( float )( v * coeff + offset ) );
            }
            return floats;
        }

        public static List<List<float>> ReadAndScaleShortsArrayWaves( BinaryReader reader, int count, double coeff, double offset, int numPointsWave )
        {
            var byteArray = reader.ReadBytes( count * numPointsWave * 2 );
            short[] values = new short[ count * numPointsWave ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            var all = new List<List<float>>();
            long index = 0;
            for ( int i = 0; i < count; i++ ) {
                var wave = new List<float>();
                for ( int j = 0; j < numPointsWave; j++ ) {
                    wave.Add( ( float )( values[ index ] * coeff + offset ) );
                    index++;
                }
                all.Add( wave );
            }
            return all;
        }

        public static List<List<float>> ReadFloatArrayWaves( BinaryReader reader, int count, int numPointsWave )
        {
            var byteArray = reader.ReadBytes( count * numPointsWave * 4 );
            float[] values = new float[ count * numPointsWave ];
            Buffer.BlockCopy( byteArray, 0, values, 0, byteArray.Length );
            var all = new List<List<float>>();
            long index = 0;
            for ( int i = 0; i < count; i++ ) {
                var wave = new List<float>();
                for ( int j = 0; j < numPointsWave; j++ ) {
                    wave.Add( ( float )( values[ index ] ) );
                    index++;
                }
                all.Add( wave );
            }
            return all;
        }

        public static byte[] GetBytesFromStringUTF8( string text, int totalBytes )
        {
            byte[] bytes = new byte[ totalBytes ];
            byte[] stringBytes = System.Text.Encoding.UTF8.GetBytes( text );
            for ( int i = 0; i < totalBytes; i++ ) {
                bytes[ i ] = 0;
                if ( i < stringBytes.Length ) {
                    bytes[ i ] = stringBytes[ i ];
                }
            }
            return bytes;
        }

        public static byte[] GetBytesFromString( string text, int totalBytes )
        {
            byte[] bytes = new byte[ totalBytes ];
            byte[] stringBytes = System.Text.Encoding.ASCII.GetBytes( text );
            for ( int i = 0; i < totalBytes; i++ ) {
                bytes[ i ] = 0;
                if ( i < stringBytes.Length ) {
                    bytes[ i ] = stringBytes[ i ];
                }
            }
            return bytes;
        }


        public static void WriteListOfNumbers( BinaryWriter writer, List<long> theList )
        {
            var byteArray = new byte[ theList.Count * 8 ];
            Buffer.BlockCopy( theList.ToArray(), 0, byteArray, 0, byteArray.Length );
            writer.Write( byteArray );
        }

        public static void WriteListOfNumbers( BinaryWriter writer, List<uint> theList )
        {
            var byteArray = new byte[ theList.Count * 4 ];
            Buffer.BlockCopy( theList.ToArray(), 0, byteArray, 0, byteArray.Length );
            writer.Write( byteArray );
        }

        public static void WriteListOfNumbers( BinaryWriter writer, List<int> theList )
        {
            var byteArray = new byte[ theList.Count * 4 ];
            Buffer.BlockCopy( theList.ToArray(), 0, byteArray, 0, byteArray.Length );
            writer.Write( byteArray );
        }

        public static void WriteListOfNumbers( BinaryWriter writer, List<short> theList )
        {
            var byteArray = new byte[ theList.Count * 2 ];
            Buffer.BlockCopy( theList.ToArray(), 0, byteArray, 0, byteArray.Length );
            writer.Write( byteArray );
        }


        public static void WriteListOfNumbers( BinaryWriter writer, List<float> theList )
        {
            var byteArray = new byte[ theList.Count * 4 ];
            Buffer.BlockCopy( theList.ToArray(), 0, byteArray, 0, byteArray.Length );
            writer.Write( byteArray );
        }

        public static double CalcScaleFloatsToShorts( List<float> numbers )
        {
            var max = numbers.Max();
            var min = numbers.Min();
            double absMax = Math.Max( Math.Abs( min ), Math.Abs( max ) );
            if ( absMax == 0.0 ) {
                return 1.0;
            } else {
                return 32767.0 / absMax;
            }
        }

        public static double CalcScaleFloatsToShorts( List<List<float>> numbers )
        {
            var maxes = new List<float>();
            var mins = new List<float>();
            foreach ( var wf in numbers ) {
                maxes.Add(wf.Max());
                mins.Add(wf.Min());
            }
            var max = maxes.Max();
            var min = mins.Min();
            double absMax = Math.Max( Math.Abs( min ), Math.Abs( max ) );
            if ( absMax == 0 ) {
                return 1.0;
            } else {
                return 32767.0 / absMax;
            }
        }

        public static List<short> ConvertContValuesToShorts( Continuous cont )
        {
            var res = new List<short>();
            foreach ( var v in cont.Values ) {
                res.Add( ( short )( v * cont.CalulatedCoeffToScaleFloatsToShorts ) );
            }
            return res;
        }

        public static List<short> ConvertWaveValuesToShorts( Waveform wave )
        {
            var res = new List<short>();
            foreach ( var wf in wave.Values ) {
                foreach ( var v in wf ) {
                    res.Add( ( short )( v * wave.CalulatedCoeffToScaleFloatsToShorts ) );
                }
            }
            return res;
        }
    }
}
